# -*- encoding: utf-8 -*-
"""
dws.core.didding module

"""

import itertools
import json
import math
import re
import urllib.parse
from base64 import urlsafe_b64encode
from functools import reduce

from keri import kering
from keri.app import habbing
from keri.core import coring
from keri.core.eventing import Kever, Kevery
from keri.db.basing import Baser
from keri.help import helping
from keri.vdr import credentialing

from dws import DidWebsError, UnknownAID, log_name, ogler
from dws.core import habs, oobiing

logger = ogler.getLogger(log_name)

DID_KERI_RE = re.compile(
    pattern=r'\Adid:keri:'
    r'(?:(?P<aid>[^:?]+))'
    r'(?P<query>\?.*)?\Z',
    flags=re.IGNORECASE,
)
DID_WEBS_RE = re.compile(
    pattern=r'\Adid:web(s)?:'
    r'(?P<domain>[^%:]+)'
    r'(?:%3a(?P<port>\d+))?'
    r'(?::(?P<path>.+?))?'
    r'(?::(?P<aid>[^:?]+))'
    r'(?P<query>\?.*)?\Z',
    flags=re.IGNORECASE,
)
DID_WEBS_UNENCODED_PORT_RE = re.compile(
    pattern=r'\Adid:web(s)?:'
    r'(?P<domain>[^%:]+)'
    r'(?::(?P<port>\d+))?'
    r'(?::(?P<path>.+?))?'
    r'(?::(?P<aid>[^:?]+))'
    r'(?P<query>\?.*)?\Z',
    flags=re.IGNORECASE,
)

# Detects the malformed shape:
#   did:webs:<domain>:<port>[?<query>]
# where <port> is unencoded (":" not "%3A") and the required AID segment is absent.
NO_AID_UNENCODED_PORT_RE = re.compile(
    pattern=r'\Adid:web(s)?:'
    r'(?P<domain>[^%:]+)'
    r':(?P<port>\d+)'
    r'(?P<query>\?.*)?\Z',
    flags=re.IGNORECASE,
)

DID_TIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
DID_TIME_PATTERN = re.compile(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z')

DES_ALIASES_SCHEMA = 'EN6Oh5XSD5_q2Hgu-aqpdfbVepdpYpFlgz6zvJL5b_r5'

DID_RES_META_FIELD = 'didResolutionMetadata'
DD_META_FIELD = 'didDocumentMetadata'
DD_FIELD = 'didDocument'
VMETH_FIELD = 'verificationMethod'


def parse_did_keri(did):
    """
    Parse a did:keri DID with regex to return the AID

    Returns:
        str: AID extracted from the did:keri DID
    """
    match = DID_KERI_RE.match(did)
    if match is None:
        raise ValueError(f'{did} is not a valid did:keri DID')

    aid = match.group('aid')
    query = match.group('query')

    try:
        _ = coring.Prefixer(qb64=aid)
    except Exception as e:
        raise ValueError(f'{aid} is an invalid AID')

    return aid, query


def parse_did_webs(did: str):
    """
    Parse a did:webs DID with regex to return the domain, port, path, and AID

    Returns:
        (str, str, str, str, str): domain, port, path, AID
    """
    match = DID_WEBS_RE.match(did)
    if match is None:
        raise ValueError(f'{did} is not a valid did:web(s) DID')

    domain, port, path, aid, query = match.group('domain', 'port', 'path', 'aid', 'query')

    try:
        _ = coring.Prefixer(qb64=aid)
    except Exception as e:
        raise ValueError(f'{aid} is an invalid AID')

    return domain, port, path, aid, query


def parse_query_string(query: str) -> dict:
    if not query or query == '?':
        return {}
    query = query.lstrip('?')
    parsed = urllib.parse.parse_qs(query)
    result = {}
    for key, values in parsed.items():
        value = values[0] if values else ''
        if value.lower() == 'true':
            result[key] = True
        elif value.lower() == 'false':
            result[key] = False
        else:
            try:
                result[key] = int(value)
            except ValueError:
                result[key] = value
    return result


def re_encode_invalid_did_webs(did: str):
    """
    Normalizes did:webs strings that arrive with an unencoded domain/port separator (:<port>) by
    re-encoding it as %3A, while preserving path, AID, and query. This compensates for URL path
    quote/unquote handling in the Falcon/HIO WSGI request path.
    """
    if NO_AID_UNENCODED_PORT_RE.match(did):
        raise ValueError(f'{did} is missing an AID')

    match = DID_WEBS_UNENCODED_PORT_RE.match(did)
    if match is None:
        match = DID_WEBS_RE.match(did)
        if match is not None:
            return did  # already encoded correctly
        raise ValueError(f'{did} is not an invalidly encoded did:web(s) DID')

    domain, port, path, aid, query = match.group('domain', 'port', 'path', 'aid', 'query')

    try:
        _ = coring.Prefixer(qb64=aid)
    except Exception as e:
        raise ValueError(f'{aid} is an invalid AID')

    encoded = f'did:webs:{domain}'
    if port:
        encoded += f'%3A{port}'
    if path:
        encoded += f':{path}'
    # always add the AID
    encoded += f':{aid}'
    if query:
        encoded += f'{query}'
    return encoded


def requote(encoded_did: str):
    """
    Due to compliance with PEP3333 in PR 38 to HIO (https://github.com/ioflo/hio/pull/38) the
    WSGI container for the Falcon server URL encodes the URL path which means that  the did:webs and did:keri
    DIDs must be URL decoded, broken apart, and then re-encoded to ensure they are valid for resolution.

    This specific attribute of the WSGI environment uses urllib.parse.quote to encode the path, so we need to decode it
    environ['PATH_INFO'] = quote(requestant.path)
    """
    if encoded_did.lower().startswith('did%3awebs') or encoded_did.lower().startswith('did%3akeri'):
        # The DID is incorrectly fully URL-encoded, happens with some WSGI servers, so must decode and re-encode it
        # this happens in Falcon
        did = urllib.parse.unquote(encoded_did)
        did = re_encode_invalid_did(did)
        return did
    else:
        return encoded_did


def re_encode_invalid_did(did: str):
    """
    Parse a did:webs DID with regex to return the domain, port, path, and AID.
    This version does not URL-encode the port.

    Returns:
        (str, str, str, str): domain, port, path, AID
    """
    if did.startswith('did:webs:'):
        return re_encode_invalid_did_webs(did)
    elif did.startswith('did:keri:'):  # included for completion's sake and uniformity when parsing DIDs
        aid, query = parse_did_keri(did)
        return f'did:keri:{aid}{query if query else ""}'
    else:
        raise ValueError(f'{did} is not a valid did:webs or did:keri DID')


def generate_json_web_key_vm(pubkey, did, kid, x):
    """
    Generate a JSON Web Key (JWK) verification method for a given public key.

    Parameters:
        pubkey (str): The public key identifier (e.g., a Verfer's qb64).
        did (str): The DID to associate with the verification method.
        kid (str): The key ID for the JWK.
        x (str): The base64url-encoded public key value.
    """
    return dict(
        id=f'#{pubkey}',
        type='JsonWebKey',
        controller=strip_query(did),
        publicKeyJwk=dict(kid=f'{kid}', kty='OKP', crv='Ed25519', x=f'{x}'),
    )


def strip_query(did: str):
    if did.startswith('did:webs:') or did.startswith('did:web:'):
        domain, port, path, aid, query = parse_did_webs(did=did)
        if query is None or query == '':
            return did
        return f'did:webs:{domain}%3A{port}:{path}:{aid}'
    else:
        return did  # for did:keri


def generate_verification_methods(verfers, thold, did, aid):
    """
    Generate a verification method for each public key (Verfer) from the source key state.
    Multiple verfers implies a multisig DID, a single verfer implies a single key DID.

    Parameters:
        verfers (list[core.Verfer]): A list of Verfer instances representing the public keys.
        thold (int or list): The signing threshold, possibly multisig. If an integer, it indicates a simple multisig DID.
        did (str): The DID to associate with the verification methods.
        aid (str): The AID to associate with the verification methods.

    Returns:
        list: A list of verification methods in the format required for a DID document.
    """
    # for each public key (Verfer) in the Kever, generate a verification method
    vms = []
    for idx, verfer in enumerate(verfers):
        kid = verfer.qb64
        x = urlsafe_b64encode(verfer.raw).rstrip(b'=').decode('utf-8')
        vms.append(generate_json_web_key_vm(kid, did, kid, x))

    # Handle multi-key or multisig AID cases
    if isinstance(thold, int):
        if thold > 1:
            conditions = [vm.get('id') for vm in vms]
            vms.append(generate_threshold_proof2022(aid, did, thold, conditions))
    elif isinstance(thold, list):
        vms.append(generate_weighted_threshold_proof(thold, verfers, vms, did, aid))
    return vms


def generate_threshold_proof2022(aid, did, thold, conditions):
    """
    Generate a ConditionalProof2022 verification method for a multisig DID.

    Parameters:
        aid (str): The controlling AID to associate with the conditional proof.
        did (str): The DID to associate with the conditional proof.
        thold (int): The multisig signing threshold.
        conditions (list): List of condition verification method IDs.

    Returns:
        dict: A ConditionalProof2022 verification method
    """
    return dict(
        id=f'#{aid}',
        type='ConditionalProof2022',
        controller=strip_query(did),
        threshold=thold,
        conditionThreshold=conditions,
    )


def generate_weighted_threshold_proof2022(aid, did, threshold, conditions):
    """
    Generate a ConditionalProof2022 verification method for a multisig DID with weighted conditions.

    Parameters:
        aid (str): The controlling AID to associate with the conditional proof.
        did (str): The DID to associate with the conditional proof.
        threshold (float): The multisig signing threshold.
        conditions (list): List of condition verification method IDs with weights.

    Returns:
        dict: A ConditionalProof2022 verification method with weighted conditions.
    """
    return dict(
        id=f'#{aid}',
        type='ConditionalProof2022',
        controller=strip_query(did),
        threshold=threshold,
        conditionWeightedThreshold=conditions,
    )


def generate_weighted_threshold_proof(thold, verfers, vms, did, aid):
    """
    Compute the weighted threshold proof for a multisig DID based on the provided fraction threshold
     weights and public keys (Verfers).

    Parameters:
        thold (list): A list of fractions representing the threshold weights.
        verfers (list[core.Verfer]): A list of Verfer instances representing the public keys.
        vms (list): A list of verification methods already generated for the public keys.
        did (str): The DID to associate with the weighted threshold proof.
        aid (str): The controlling AID to associate with the weighted threshold proof.
    """
    lcd = int(math.lcm(*[fr.denominator for fr in thold[0]]))
    threshold = float(lcd / 2)
    numerators = [int(fr.numerator * lcd / fr.denominator) for fr in thold[0]]
    conditions = []
    for idx, verfer in enumerate(verfers):
        conditions.append(dict(condition=vms[idx]['id'], weight=numerators[idx]))
    return generate_weighted_threshold_proof2022(aid, did, threshold, conditions)


def gen_did_document(did, verification_methods, service_endpoints, also_known_as):
    """
    Generate a basic DID document structure.

    DID document properties:
    - id: The DID itself
    - verificationMethod: A list of verification methods derived from the Kever's verfers
    - service: A list of service endpoints derived from the hab's fetchRoleUrls and fetchWitnessUrls methods
    - alsoKnownAs: A list of designated aliases for the AID

    Parameters:
        did (str): The DID to include in the document.
        verification_methods (list): A list of verification methods.
        service_endpoints (list): A list of service endpoints.
        also_known_as (list): A list of alternative identifiers.

    Returns:
        dict: A basic DID document structure.
    """
    return dict(id=did, verificationMethod=verification_methods, service=service_endpoints, alsoKnownAs=also_known_as)


def gen_did_resolution_result(did_doc, witness_list, seq_no, equivalent_ids):
    """
    Generate a DID resolution result structure.

    Parameters:
        did_doc (dict): The DID document to include in the resolution result.
        witness_list (list): A list of witnesses AIDs
        seq_no (int): The sequence number of the latest KEL event for the AID generating the DID document.
        equivalent_ids (list): A list of equivalent IDs.

    Returns:
        dict: A DID resolution result structure containing the DID document, resolution metadata, and document metadata.
    """
    return dict(
        didDocument=did_doc,
        didResolutionMetadata=dict(contentType='application/did+json', retrieved=helping.nowUTC().strftime(DID_TIME_FORMAT)),
        didDocumentMetadata=dict(
            witnesses=witness_list,
            versionId=f'{seq_no}',
            equivalentId=equivalent_ids,
        ),
    )


def get_witness_list(baser: Baser, kever: Kever):
    witness_list = []
    for idx, eid in enumerate(kever.wits):
        for (tid, scheme), loc in baser.locs.getItemIter(keys=(eid,)):
            witness_list.append(dict(idx=str(idx), scheme=scheme, url=loc.url))
    return witness_list


def get_equiv_aka_ids(did: str, aid: str, hby: habbing.Habery, rgy: credentialing.Regery):
    equiv_ids = []
    aka_ids = []
    if did.startswith('did:webs') or did.startswith('did:web'):
        for s in gen_designated_aliases(hby, rgy, aid):
            if s.startswith('did:webs'):
                equiv_ids.append(s)
            aka_ids.append(s)
    return equiv_ids, aka_ids


def generate_did_doc(hby: habbing.Habery, rgy: credentialing.Regery, did, aid, meta=False):
    """
    Generates a DID document for the given DID and AID.

    The DID document will have one of the following structures:
    - If `meta` is True:
      - didDocument: The DID document itself (see genDidDocument for structure)
      - didResolutionMetadata: Metadata about the DID resolution process
      - didDocumentMetadata: Additional metadata about the DID document
    if `meta` is False:
    - didDocument: The DID document itself (see genDidDocument for structure)

    Parameters:
        hby (habbing.Habery): The habery instance containing the necessary data.
        rgy (credentialing.Regery): The Regery instance for the credential and registry data.
        did (str): The DID to generate the document for.
        aid (str): The AID associated with the DID.
        meta (bool, optional): If True, include metadata in the response. Defaults to False.

    Returns:
        dict of DID document structure; DID document, metadata and resolution metadata or just the DID document
    """
    if did.startswith('did:webs') or did.startswith('did:web'):
        _domain, _port, _path, parsed_aid, _query = parse_did_webs(did=did)
        if (did and aid) and parsed_aid != aid:
            raise ValueError(f'{did} does not contain AID {aid}')
    if did.startswith('did:keri'):
        aid, query = parse_did_keri(did=did)
        did = f'did:keri:{aid}'  # strip of query for did doc generation as is not needed. OOBI should be resolved by now.
    logger.debug(f'Generating DID document for\n\t{did}\nwith aid\n\t{aid}\nand metadata\n\t{meta}')

    hab: habbing.Hab | None = None
    if aid in hby.habs:
        hab = hby.habs[aid]

    kever = None
    if aid in hby.kevers:
        kever = hby.kevers[aid]
    else:
        raise UnknownAID(aid, did)

    verification_methods = generate_verification_methods(kever.verfers, kever.tholder.thold, did, aid)
    service_endpoints = gen_service_endpoints(hby, hab, kever, aid)
    equiv_ids, aka_ids = get_equiv_aka_ids(did, aid, hby, rgy)

    did_doc = gen_did_document(did, verification_methods, service_endpoints, aka_ids)
    if meta is True:
        return gen_did_resolution_result(
            did_doc=did_doc,
            witness_list=get_witness_list(hby.db, kever),
            seq_no=kever.sner.num,
            equivalent_ids=equiv_ids,
        )
    else:
        return did_doc


def to_did_web(diddoc: dict, meta=False):
    """
    Convert DID schemes for did.json DID document from did:webs did:web.

    If metadata is present then the didDocument field is replaced with the converted DID document.
    """
    if not diddoc:
        raise DidWebsError('Cannot convert empty diddoc to did:web')
    if meta:
        replaced = diddoc_to_did_web(diddoc[DD_FIELD])
        diddoc[DD_FIELD] = replaced
        return diddoc
    else:
        return diddoc_to_did_web(diddoc)


def diddoc_to_did_web(diddoc: dict):
    """Converts all did:webs DIDs in the 'id' property and verification method 'controller' properties to did:web"""
    diddoc['id'] = diddoc['id'].replace('did:webs', 'did:web')
    for verificationMethod in diddoc['verificationMethod']:
        verificationMethod['controller'] = verificationMethod['controller'].replace('did:webs', 'did:web')
    return diddoc


def diddoc_to_did_webs(diddoc: dict):
    """Converts all did:web DIDs in the 'id' property and verification method 'controller' properties to did:webs"""
    # Apply the replacement only if necessary
    if 'did:web' in diddoc['id'] and 'did:webs' not in diddoc['id']:
        diddoc['id'] = diddoc['id'].replace('did:web', 'did:webs')
        logger.debug(f'Updated id in fromDidWeb: {diddoc["id"]}')

    for verificationMethod in diddoc['verificationMethod']:
        if 'did:web' in verificationMethod['controller'] and 'did:webs' not in verificationMethod['controller']:
            verificationMethod['controller'] = verificationMethod['controller'].replace('did:web', 'did:webs')
            logger.debug(f'Updated controller in fromDidWeb: {verificationMethod["controller"]}')

    return diddoc


def from_did_web(did_json: dict, meta: bool = False):
    """
    Convert DID schemes in did.json DID document from did:web to did:webs.

    If metadata is present then the didDocument field is replaced with the converted DID document.
    """
    # Log the original state of the DID and controller
    if meta and DD_FIELD not in did_json:
        logger.debug(f'DID resolution metadata did not contain {DD_FIELD}:\n{json.dumps(did_json, indent=2)}')
        raise ValueError(f"Expected '{DD_FIELD}' in did.json when indicating resolution metadata in use.")
    diddoc = did_json[DD_FIELD] if meta else did_json
    initial_controller = diddoc['verificationMethod'][0]['controller']
    # id = diddoc["id"] if not meta else did_json[DD_FIELD]["id"]
    if not meta:
        converted_did_doc = diddoc_to_did_webs(diddoc)
        diddoc = converted_did_doc
    else:
        initial_controller = diddoc['verificationMethod'][0]['controller']
        converted_did_doc = diddoc_to_did_webs(diddoc)
        did_json[DD_FIELD] = converted_did_doc
        diddoc = did_json
    return diddoc


def extract_desg_alias_from_cred(cred: dict) -> str | None:
    """
    Extract the 'ids' property from the designated aliases ACDC credential.
    This is specific to the current designated aliases schema.
    """
    sad = cred['sad']
    status = cred['status']
    if status['et'] == 'iss' or status['et'] == 'bis':
        return sad['a']['ids']
    return None


def gen_designated_aliases(hby: habbing.Habery, rgy: credentialing.Regery, aid: str, schema: str = DES_ALIASES_SCHEMA):
    """
    Searches the entire Regery database for non-revoked, self-attested designated alias ACDCs by schema and
    returns a list of designated alias IDs using their `a.ids` field.

    Parameters:
        hby (habbing.Habery): The Habery instance containing the Regery.
        rgy (credentialing.Regery): The Regery instance for credential and registry data.
        aid (str): The AID prefix to retrieve the ACDCs for.
        schema (str): The schema to use to select the target ACDC from the local registry. Default is DES_ALIASES_SCHEMA.

    Returns:
        list: A list of designated alias IDs (a.ids) from self-attested ACDCs.
    """
    da_ids = []
    saids = rgy.reger.issus.get(keys=aid)
    scads = rgy.reger.schms.get(keys=schema)  # get credentials by schema
    # self-attested schema, and scehma is designated aliases
    saids = [saider for saider in saids if saider.qb64 in [saider.qb64 for saider in scads]]
    if not saids:
        return da_ids
    if aid in hby.habs:
        # if aid is in a hab in the local habery then get self issued ACDCS
        creds = rgy.reger.cloneCreds(saids, hby.habs[aid].db)
    else:
        # otherwise if AID is just in kevers and is non-local, get creds issued by the AID in the DID
        creds = rgy.reger.cloneCreds(saids, hby.db)
    da_ids = [extract_desg_alias_from_cred(cred) for _, cred in enumerate(creds)]
    return list(itertools.chain.from_iterable(da_ids))


def gen_service_endpoints(hby: habbing.Habery, hab: habbing.Hab, kever: Kever, aid: str):
    """Generate service endpoints including both witness and delegation service endpoints."""
    serv_ends = []
    # TODO: move Location Scheme and Endpoint Role Authorization to dedicated function
    if hab and hasattr(hab, 'fetchRoleUrls'):
        ends = hab.fetchRoleUrls(cid=aid)
        serv_ends.extend(add_ends(ends))
        ends = hab.fetchWitnessUrls(cid=aid)
        serv_ends.extend(add_ends(ends))
    else:
        ends = habs.get_role_urls(baser=hby.db, kever=kever)
        serv_ends.extend(add_ends(ends))

    if hasattr(hab, 'delpre') and hab.delpre is not None:
        del_serv_end = gen_delegation_service(hby=hby, pre=hab.pre, delpre=hab.delpre)
        serv_ends.extend(del_serv_end)
    return serv_ends


def add_ends(ends):
    def process_role(role):
        return reduce(lambda rs, eids: rs + process_eids(eids, role), ends.getall(role), [])

    def process_eids(eids, role):
        return reduce(lambda es, eid: es + process_eid(eid, eids[eid], role), eids, [])

    def process_eid(eid, val, role):
        v = dict(id=f'#{eid}/{role}', type=role, serviceEndpoint={proto: f'{host}' for proto, host in val.items()})
        return [v]

    return reduce(lambda emit, role: emit + process_role(role), ends, [])


def gen_delegation_service(hby: habbing.Habery, pre: str, delpre: str):
    """Returns an array of delegation service endpoints for the delegator AID."""
    seal = dict(i=pre, s='0', d=pre)
    dserder = hby.db.fetchLastSealingEventByEventSeal(pre=delpre, seal=seal)
    approval_evt_digest = dserder.sad['a'][0]['d']
    del_oobi = oobiing.get_resolved_oobi(hby=hby, pre=delpre)
    if del_oobi is None:
        logger.error(f'No resolved OOBI found for delegate AID {delpre} for delegator AID {pre}')
        return []
    return [dict(id=approval_evt_digest, type='DelegatorOOBI', serviceEndpoint=del_oobi)]



import argparse

from hio.base import doing

import dws


def create_parser() -> argparse.ArgumentParser:
    """
    Create and return the argument parser for the version command.
    """
    parser = argparse.ArgumentParser(description='Print version of did:webs resolver')
    parser.set_defaults(handler=lambda args: handler(args))
    parser.add_argument(
        '--verbose', '-v', help='verbose version information', required=False, default=False, action='store_true'
    )
    return parser


parser = create_parser()


def handler(args: argparse.Namespace) -> list[doing.Doer]:
    kwa = dict(args=args)
    return [doing.doify(version, **kwa)]


def version(tymth, tock=0.0, existing=None, **opts):
    """Prints the version of the library"""
    _ = yield tock
    args = opts['args']
    verbose = args.verbose
    if not verbose:
        print(f'{dws.__version__}')
    else:
        print(f'did:webs resolver version: {dws.__version__}')



# -*- encoding: utf-8 -*-
"""
dws.core.ends.keri_cesr_resource_end module

"""

import falcon
from keri.app import habbing
from keri.app.habbing import Hab, Habery
from keri.vdr import credentialing
from keri.vdr.viring import Reger

from dws.core import artifacting

KERI_CESR = 'keri.cesr'
CESR_MIME = 'application/cesr'


class KeriCesrResourceEnd:
    """
    keri.cesr resource endpoint for accessing all KEL, TEL, and ACDC artifacts needed for did:webs DIDs
    """

    def __init__(self, hby: habbing.Habery, rgy: credentialing.Regery):
        """
        Initialize did:webs keri.cesr artifact endpoint that will pull designated aliases from the specified registry.

        Parameters:
            hby (Habery): Database environment for AIDs to expose
            rgy (Regery): Registry for credential and registry data
        """
        self.hby: habbing.Habery = hby
        self.rgy: credentialing.Regery = rgy

    def on_get(self, req, rep, aid):
        """GET endpoint for accessing {KERI_CESR} stream for AID

        Parameters:
            req (Request) Falcon HTTP Request object:
            rep (Response) Falcon HTTP Response object:
            aid (str): AID to access {KERI_CESR} stream for
        """
        # Read the DID from the parameter extracted from path or manually extract
        if not req.path.endswith(f'/{KERI_CESR}'):
            raise falcon.HTTPBadRequest(description=f'invalid {KERI_CESR} DID URL {req.path}')

        if aid not in self.hby.kevers:
            raise falcon.HTTPNotFound(description=f'KERI AID {aid} not found')

        hab = self.hby.habs[aid]
        keri_cesr = gen_keri_cesr(hab, self.rgy.reger, aid)

        rep.status = falcon.HTTP_200
        rep.content_type = CESR_MIME
        rep.data = keri_cesr


def gen_keri_cesr(hab: Hab, reger: Reger, aid: str) -> bytearray:
    """Load KEL, TEL, and ACDC CESR bytes for the givne AID."""
    keri_cesr = bytearray()
    keri_cesr.extend(artifacting.gen_kel_cesr(hab, pre=aid))  # add KEL CESR stream
    keri_cesr.extend(artifacting.gen_loc_schemes_cesr(hab, aid))
    keri_cesr.extend(artifacting.gen_des_aliases_cesr(hab, reger, aid))  # add designated aliases TELs and ACDCs
    return keri_cesr



# -*- encoding: utf-8 -*-
"""
dws.app.cli.commands.did.webs.service module

"""

import argparse
from typing import List

from hio.base import Doer
from keri.app import oobiing
from keri.vdr import credentialing

from dws import log_name, ogler, set_log_level
from dws.core import artifacting, habs

parser = argparse.ArgumentParser(description='Launch web server capable of serving KERI AIDs as did:webs and did:web DIDs')
parser.set_defaults(handler=lambda args: launch(args), transferable=True)
parser.add_argument(
    '-d',
    '--did-path',
    action='store',
    default='',
    help="did:webs path segment in URL format between {host}%3A{port} and {aid}. Example: 'somepath/somesubpath'",
)
parser.add_argument('-p', '--http', action='store', default=7676, help='Port on which to listen for did:webs requests')
parser.add_argument('-n', '--name', action='store', required=True, help='Name of controller.')
parser.add_argument('-a', '--alias', action='store', required=True, help='Alias of controller.')
parser.add_argument(
    '--base', '-b', help='additional optional prefix to file location of KERI keystore', required=False, default=''
)
parser.add_argument(
    '--passcode', help='22 character encryption passcode for keystore (is not saved)', dest='bran', default=None
)  # passcode => bran
parser.add_argument('--config-dir', '-c', dest='config_dir', help='directory override for configuration data', default=None)
parser.add_argument('--config-file', dest='config_file', action='store', help='configuration filename override')
parser.add_argument(
    '-m',
    '--meta',
    action='store_true',
    required=False,
    default=False,
    help='Whether to include metadata (True), or only return the DID document (False)',
)
parser.add_argument('--keypath', action='store', required=False, default=None)
parser.add_argument('--certpath', action='store', required=False, default=None)
parser.add_argument('--cafilepath', action='store', required=False, default=None)
parser.add_argument(
    '--loglevel',
    action='store',
    required=False,
    default='CRITICAL',
    help='Set log level to DEBUG | INFO | WARNING | ERROR | CRITICAL. Default is CRITICAL',
)

logger = ogler.getLogger(log_name)


def launch(args):
    """Handle CLI command for serving did:webs artifacts."""
    set_log_level(args.loglevel, logger)
    http_port = args.http
    try:
        http_port = int(http_port)
    except ValueError:
        logger.error(f'Invalid port number: {http_port}. Must be an integer.')
        raise

    logger.info(f'Launched did:webs artifact webserver: {http_port}')
    return create_artifact_server_doers(
        name=args.name,
        base=args.base,
        bran=args.bran,
        config_file=args.config_file,
        config_dir=args.config_dir,
        alias=args.alias,
        meta=args.meta,
        did_path=args.did_path,
        http_port=http_port,
        keypath=args.keypath,
        certpath=args.certpath,
        cafilepath=args.cafilepath,
    )


def create_artifact_server_doers(
    name: str,
    base: str,
    bran: str,
    config_file: str,
    config_dir: str,
    alias: str,
    meta: bool,
    did_path: str,
    http_port: int,
    keypath: str = None,
    certpath: str = None,
    cafilepath: str = None,
) -> List[Doer]:
    """Create a list of Doers for serving did:webs artifacts."""
    cf = habs.get_habery_configer(name=config_file, base=base, head_dir_path=config_dir)
    hby, hby_doer = habs.get_habery_and_doer(name, base, bran, cf)
    rgy = credentialing.Regery(hby=hby, name=hby.name, base=hby.base, temp=hby.temp)
    oobiery = oobiing.Oobiery(hby=hby)
    doers = oobiery.doers + [hby_doer]

    doers += artifacting.dyn_artifact_svr_doers(
        hby=hby,
        rgy=rgy,
        alias=alias,
        http_port=http_port,
        did_path=did_path,
        meta=meta,
        keypath=keypath,
        certpath=certpath,
        cafilepath=cafilepath,
    )

    return doers



# -*- encoding: utf-8 -*-
"""
dws.core.resolving module

"""

import io
import json
import logging
import os
from typing import Callable, List

import falcon
from falcon import media
from hio.base import doing
from hio.core import http, tcp
from keri import kering
from keri.app import habbing, oobiing
from keri.app.habbing import Habery
from keri.app.oobiing import Oobiery
from keri.core import eventing, routing
from keri.db import basing
from keri.help import helping
from keri.peer import exchanging
from keri.vdr import credentialing, verifying
from keri.vdr import eventing as teventing
from keri.vdr.credentialing import Regery

from dws import ArtifactResolveError, log_name, ogler
from dws.core import didding, ends, requesting
from dws.core.requesting import load_url_with_requests

logger = ogler.getLogger(log_name)


def gen_dws_urls(did: str) -> (str, str, str):
    """
    Generate the did.json and keri.cesr HTTP URLs for a did:webs DID.

    Parameters:
        did (str): The did:webs DID to generate URLs for.

    Returns:
          (str, str, str): the AID, did.json, and keri.cesr URLs parsed from the DID
    """
    domain, port, path, aid, query = didding.parse_did_webs(did=did)

    opt_port = f':{port}' if port is not None else ''
    opt_path = f'/{path.replace(":", "/")}' if path is not None else ''
    base_url = f'https://{domain}{opt_port}{opt_path}/{aid}'

    # did.json for DID Document
    dd_url = f'{base_url}/{ends.DID_JSON}'
    if query:
        dd_url += f'{query}'

    # keri.cesr for CESR stream
    kc_url = f'{base_url}/{ends.KERI_CESR}'
    return aid, dd_url, kc_url


def get_dws_artifacts(did: str, load_url: Callable = load_url_with_requests, timeout: float = 5.0) -> (str, bytes, bytes):
    """
    Execute HTTP calls for the did.json and keri.cesr artifacts for a did:webs DID.

    Parameters:
        did (str): The did:webs DID to resolve.
        load_url (Callable): Function to load URLs, can be mocked for testing.
        timeout (float): Timeout for HTTP requests in seconds.

    Returns:
        (str, bytes, bytes): The AID, did.json content, and keri.cesr content for the given did:webs DID.
    """
    aid, dd_url, kc_url = gen_dws_urls(did=did)

    # Load the did doc
    logger.info(f'Loading DID Doc from {dd_url}')
    dd_res = load_url(dd_url, timeout=timeout)
    logger.debug(f'Got DID doc: {dd_res.decode("utf-8")}')

    # Load the KERI CESR
    logger.info(f'Loading KERI CESR from {kc_url}')
    kc_res = load_url(kc_url, timeout=timeout)
    logger.debug(f'Got KERI CESR: {kc_res.decode("utf-8")}')

    return aid, dd_res, kc_res


def save_cesr(hby: Habery, rgy: Regery, kc_res: bytes, aid: str = None):
    """
    Save the resolved keri.cesr stream to the local keystore by parsing the CESR objects in it
    with the appropriate message processor. This makes the designated aliases ACDC and the location
    scheme records available to use for generating the correct did.json document during did doc verification.

    Parameters:
        hby (Habery): The Habery instance containing the KERI database.
        rgy (Regery): The Regery instance for credential and registry management.
        kc_res (bytes): The KERI CESR stream to save.
        aid (str): The AID of the identifier to which the CESR belongs. If None, it will be derived from the CESR.
    """
    logger.debug('Saving KERI CESR to hby: %s', kc_res.decode('utf-8'))
    # CESR message handlers
    rtr = routing.Router()
    rvy = routing.Revery(db=hby.db, rtr=rtr)  # Have to create the Revery before Kevery so the Kevery.kvr is set.
    exc = exchanging.Exchanger(hby=hby, handlers=[])
    kvy = eventing.Kevery(db=hby.db, rvy=rvy)
    tvy = teventing.Tevery(db=hby.db, reger=rgy.reger)
    vry = verifying.Verifier(hby=hby, reger=rgy.reger)
    kvy.registerReplyRoutes(router=rtr)  # make sure LocScheme and EndRole records are processed
    tvy.registerReplyRoutes(router=rtr)  # make sure ACDC rpy messages are processed
    local = True if aid in hby.habs else False

    # Call to parse the stream.
    hby.psr.parse(ims=bytearray(kc_res), kvy=kvy, tvy=tvy, vry=vry, rvy=rvy, exc=exc, local=local)

    # After parsing then the AID should be in kevers, meaning the KEL for the AID is locally available
    if aid not in hby.kevers:
        raise kering.KeriError(f'KERI CESR parsing and saving failed, KERI AID {aid} not found in habery')

    # Process escrows to ensure all events are processed
    kvy.processEscrows()
    tvy.processEscrows()
    exc.processEscrow()
    vry.processEscrows()
    rvy.processEscrowReply()


def get_generated_did_doc(
    hby: habbing.Habery,
    rgy: credentialing.Regery,
    did: str,
    meta: bool,
):
    """
    Get the did:webs DID document generated from the did.json and keri.cesr files loaded for a given did:webs DID.
    If metadata is requested then the response will include DID resolution metadata.

    Parameters:
        hby (habbing.Habery): The Habery instance containing the KERI database.
        rgy (credentialing.Regery): The Regery instance for credential and registry management.
        did (str): The did:webs DID to generate the document for.
        meta (bool): Whether to include metadata in the DID document.
    """
    aid, dd_url, kc_url = gen_dws_urls(did=did)
    dd = didding.generate_did_doc(hby, rgy, did=did, aid=aid, meta=meta)
    if meta:
        dd[didding.DD_META_FIELD]['didDocUrl'] = dd_url
        dd[didding.DD_META_FIELD]['keriCesrUrl'] = kc_url
    return dd


def error_resolution_response(error_message: str, differences: list) -> dict:
    """
    Generate an error response for DID resolution.
    """
    resolution = dict()  # Copy the actual DID document to modify it
    resolution[didding.DD_FIELD] = None
    resolution[didding.DID_RES_META_FIELD] = dict()
    resolution[didding.DID_RES_META_FIELD]['error'] = 'notVerified'
    resolution[didding.DID_RES_META_FIELD]['errorMessage'] = error_message
    resolution[didding.DID_RES_META_FIELD]['differences'] = differences
    return resolution


def verify(dd_expected: dict, dd_actual: dict, meta: bool = False) -> (bool, dict):
    """
    Verify the DID document against the KERI event stream.

    Returns:
         tuple(bool, dict): (verified, dd) where verified is a boolean indicating verification status
    """
    dd_exp = dd_expected
    dd_act = dd_actual
    # only compare the did document, not the DID resolution metadata
    if meta:
        dd_exp = dd_expected[didding.DD_FIELD]
        dd_act = dd_actual[didding.DD_FIELD]
    verified, differences = compare_did_docs(dd_exp, dd_act)

    if verified:
        logger.info(f'DID document verified')
        return True, dd_expected
    else:
        logger.info(f'DID document verification failed')
        return (
            False,
            error_resolution_response(
                error_message='The DID document could not be verified against the KERI event stream', differences=differences
            ),
        )


def compare_did_docs(expected, actual) -> (bool, list):
    """
    Performs did:webs DID document verification by performing a simple, deep dictionary comparison
    using Python's built-in equality operator comparison.

    Returns:
        tuple(bool, list): (verified, differences) where verified is a boolean indicating verification status
    """
    # TODO determine what to do with BADA RUN things like services (witnesses) etc.
    if (
        expected != actual
    ):  # Python != and == perform a deep object value comparison; this is not reference equality, it is value equality
        differences = diff_dicts(expected, actual)
        logger.error(f'Differences found in DID Doc verification: {differences}')
        return False, differences
    else:
        return True, []


def diff_dicts(expected, actual, path=''):
    """Recursively compare two dictionaries and return differences."""
    logger.error(f'Comparing dictionaries:\nexpected:\n{expected}\n \nand\n \nactual:\n{actual}')
    differences = []

    if isinstance(expected, dict):
        for k in expected.keys():
            # Construct current path
            current_path = f'{path}.{k}' if path else k
            logger.error(f'Comparing key {current_path}')

            # Key not present in the actual dictionary
            if k not in actual:
                differences.append((current_path, expected[k], None))
                logger.error(f'Key {current_path} not found in the actual dictionary')
                continue

            # If value in expected is a dictionary but not in actual
            if isinstance(expected[k], dict) and not isinstance(actual[k], dict):
                differences.append((current_path, expected[k], actual[k]))
                logger.error(f'{current_path} is a dictionary in expected, but not in actual')
                continue

            # If value in actual is a dictionary but not in expected
            if isinstance(actual[k], dict) and not isinstance(expected[k], dict):
                differences.append((current_path, expected[k], actual[k]))
                logger.error(f'{current_path} is a dictionary in actual, but not in expected')
                continue

            # If value is another dictionary, recurse
            if isinstance(expected[k], dict) and isinstance(actual[k], dict):
                differences.append(diff_dicts(expected[k], actual[k], current_path))
            # Compare non-dict values
            elif expected[k] != actual[k]:
                differences.append((current_path, expected[k], actual[k]))
                logger.error(f'Different values for key {current_path}: {expected[k]} (expected) vs. {actual[k]} (actual)')

        if isinstance(actual, dict):
            # Check for keys in actual that are not present in expected
            for k in actual.keys():
                current_path = f'{path}.{k}' if path else k
                if k not in expected:
                    differences.append((current_path, None, actual[k]))
                    logger.error(f'Key {current_path} not found in the expected dictionary')
        else:
            differences.append((path, expected, None))
            logger.error(f'Expecting actual did document to contain dictionary {expected}')
    elif isinstance(expected, list):
        if len(expected) != len(actual):
            differences.append((path, expected, actual))
            logger.error(f'Expected list {expected} and actual list {actual} are not the same length')
        else:
            for i in range(len(expected)):
                differences.append(diff_dicts(expected[i], actual[i], path))
    else:
        if expected != actual:
            differences.append((path, expected, actual))
            logger.error(f'Different values for key {path}: {expected} (expected) vs. {actual} (actual)')
    return differences


def wrap_metadata(did_doc: dict, did: str, aid: str, hby: habbing.Habery, rgy: credentialing.Regery) -> dict:
    # wrap the loaded DID document with resolution metadata if it is missing
    kever = hby.kevers[aid]
    equiv_ids, _ = didding.get_equiv_aka_ids(did, aid, hby, rgy)
    return didding.gen_did_resolution_result(
        did_doc=did_doc, witness_list=didding.get_witness_list(hby.db, kever), seq_no=kever.sner.num, equivalent_ids=equiv_ids
    )


def resolve(
    hby: habbing.Habery,
    rgy: credentialing.Regery,
    did: str,
    meta: bool = False,
    load_url: Callable = load_url_with_requests,
    timeout: float = 5.0,
) -> (bool, dict):
    """
    Resolve a did:webs DID and returl the verification result.

    Parameters:
        hby (habbing.Habery): The Habery instance containing the KERI database.
        rgy (credentialing.Regery): The Regery instance for credential and registry management.
        did (str): The did:webs DID to resolve.
        meta (bool): Whether to include metadata in the DID document.
        load_url (Callable): Function to load URLs, can be mocked for testing.
        timeout (float): Timeout for HTTP requests in seconds.

    Returns:
        tuple(bool, dict): (verified, resolution) where verified is a boolean indicating verification status
    """
    try:
        aid, dd_res, kc_res = get_dws_artifacts(did=did, load_url=load_url, timeout=timeout)
    except ArtifactResolveError as e:
        logger.error(f'Failed to resolve DID {did}: {e}')
        return False, {'error': str(e)}
    except Exception as e:
        logger.error(f'Unexpected error while resolving DID {did}: {e}')
        return False, {'error': f'Unexpected error while resolving DID {did}: {e}'}
    save_cesr(hby=hby, rgy=rgy, kc_res=kc_res, aid=aid)
    loaded_dd = json.loads(dd_res.decode('utf-8'))
    if meta and didding.DD_FIELD not in loaded_dd:
        loaded_dd = wrap_metadata(loaded_dd, did, aid, hby, rgy)
    dd_actual = didding.from_did_web(loaded_dd, meta)
    dd_expected = get_generated_did_doc(hby=hby, rgy=rgy, did=did, meta=meta)
    return verify(dd_expected, dd_actual, meta=meta)


def keri_headers() -> List[str]:
    """HTTP header array for both CESR content types and Signify headers."""
    return [
        'cesr-attachment',
        'cesr-date',
        'content-type',
        'signature',
        'signature-input',
        'signify-resource',
        'signify-timestamp',
    ]


def cors_middleware() -> falcon.CORSMiddleware:
    """Wide open CORS HTTP header Falcon middleware."""
    return falcon.CORSMiddleware(
        allow_origins='*',
        allow_credentials='*',
        expose_headers=keri_headers(),
    )


# noinspection PyMethodMayBeStatic
class RequestLoggerMiddleware:
    """Simple request logger Falcon middleware."""

    def process_request(self, req: falcon.Request, resp: falcon.Response):
        """Log incoming requests."""
        logger.info('Request received : %s %s', req.method, req.url)
        logger.debug('Request headers : %s', req.headers)
        if req.content_length and logger.isEnabledFor(logging.DEBUG):
            # Read and re-set the stream to allow further processing
            body = req.stream.read()
            decoded_body = body.decode('utf-8') if body else '<empty>'
            logger.debug('Request body    : %s', decoded_body)
            req.env['wsgi.input'] = io.BytesIO(body)  # Reset the stream for further processing
            req.env['CONTENT_LENGTH'] = str(len(body))  # match WSGI env content length to body length
        else:
            logger.debug('Request body    : No body')

    def process_response(self, req: falcon.Request, resp: falcon.Response, resource, req_succeeded):
        """Log outgoing responses."""
        logger.info(f'Response status  : %s on %s %s', resp.status, req.method, req.url)
        logger.debug(f'Response headers: %s', resp.headers)


def falcon_app() -> falcon.App:
    """Create a Falcon app instance with open CORS settings."""
    app = falcon.App(middleware=[cors_middleware(), RequestLoggerMiddleware()])
    app.req_options.media_handlers = media.Handlers(
        {
            'application/json': media.JSONHandler(),
            'application/did+ld+json': media.JSONHandler(),  # Map DID JSON-LD to JSON parser
            'multipart/form-data': media.MultipartFormHandler(),
            'application/x-www-form-urlencoded': media.URLEncodedFormHandler(),
        }
    )
    app.resp_options.media_handlers = media.Handlers(
        {
            'application/json': media.JSONHandler(),
            'application/did+ld+json': media.JSONHandler(),  # Ensure responses can use it
            'application/did-resolution': media.JSONHandler(),  # Map DID Resolution to JSON parser
        }
    )
    return app


def tls_falcon_server(
    app: falcon.App, http_port: int, keypath: str | None, certpath: str | None, cafilepath: str | None
) -> http.Server:
    """Add TLS support to a Falcon server."""
    if keypath is not None:
        servant = tcp.ServerTls(certify=False, keypath=keypath, certpath=certpath, cafilepath=cafilepath, port=int(http_port))
    else:
        servant = None

    server = http.Server(port=int(http_port), app=app, servant=servant)
    return server


def setup_resolver(
    hby, rgy, oobiery, http_port, static_files_dir=None, did_path=None, keypath=None, certpath=None, cafilepath=None
):
    """Setup serving package and endpoints

    Parameters:
        hby (habbing.Habery): identifier database environment
        rgy (credentialing.Regery): Doer for the identifier database environment
        oobiery (Oobiery): OOBI management environment
        http_port (int): external port to listen on for HTTP messages
        static_files_dir (str): directory to serve static files from, default is None (disabled)
        did_path (str): path segment of the did:webs URL to host the did:webs artifacts on, disabled if None
        keypath (str | None): path to the TLS private key file, default is None (disabled)
        certpath (str | None): path to the TLS certificate file, default is None (disabled)
        cafilepath (str | None): path to the CA certificate file, default is None (disabled)
    Returns:
        list: list of Doers to run in the Tymist
    """
    logger.info(f'Setting up Resolver HTTP server Doers on port {http_port}')
    app = falcon_app()

    server = tls_falcon_server(app, http_port=http_port, keypath=keypath, certpath=certpath, cafilepath=cafilepath)
    http_server_doer = http.ServerDoer(server=server)

    load_ends(app, hby=hby, rgy=rgy, oobiery=oobiery, static_files_dir=static_files_dir, did_path=did_path)

    doers = [http_server_doer]

    return doers


def get_serve_dir(static_files_dir: str | None, did_doc_dir: str):
    """
    When did DOC dir is absolute path, return it. If not, then check if static files is absolute.
    If not, then combine current working, static files, and did doc dir to get the full path.
    If static files is absolute, then combine it with did doc dir to get the full path.
    """
    if not os.path.isabs(did_doc_dir):
        if not os.path.isabs(static_files_dir):
            return os.path.join(os.getcwd(), static_files_dir, did_doc_dir)
        return os.path.join(os.path.abspath(static_files_dir), did_doc_dir)
    return did_doc_dir


def serve_artifacts(app: falcon.App, hby: habbing.Habery, static_files_dir: str | None = None, did_path: str = ''):
    """Configures a Falcon HTTP server with a static file server for did.json and keri.cesr files based on a local directory."""
    did_doc_dir = get_serve_dir(static_files_dir, hby.cf.get().get('did.doc.dir', ''))
    route = '' if did_path is None else f'/{did_path}'
    logger.info(f'Serving static files from {did_doc_dir} at route {route}')
    # Host did:webs artifacts only if static path specified
    app.add_static_route(route, did_doc_dir)


def load_ends(
    app: falcon.App,
    hby: habbing.Habery,
    rgy: credentialing.Regery,
    oobiery: oobiing.Oobiery,
    static_files_dir: str,
    did_path: str = '',
):
    """Set up Falcon HTTP server endpoints for resolving DIDs and hosting static files"""
    if static_files_dir is not None:
        serve_artifacts(app, hby, static_files_dir, did_path)
    resolve_end = UniversalResolverResource(hby=hby, rgy=rgy, oobiery=oobiery, load_url=requesting.load_url_with_requests)
    app.add_route('/1.0/identifiers/{did}', resolve_end)
    app.add_route('/health', ends.HealthEnd())


class UniversalResolverResource:
    """
    HTTP Resource enabling the Universal Resolver to resolve did:webs and did:keri DIDs using the /v1.0/identifiers/{did}  endpoint.
    """

    TimeoutArtifactResolution = 5.0  # seconds to wait for artifact resolution before timing out

    def __init__(
        self,
        hby: habbing.Habery,
        rgy: credentialing.Regery,
        oobiery: oobiing.Oobiery,
        load_url: Callable = load_url_with_requests,
    ):
        """Create Endpoints for discovery and resolution of OOBIs

        Parameters:
            hby (Habery): identifier database environment
            rgy (Regery): Credential and registry data manager
            oobiery (Oobiery): OOBI management environment
            load_url (Callable): HTTP request function to use to load did.json and keri.cesr - simplifies testing
        """
        self.hby: Habery = hby
        self.rgy = (
            rgy if rgy else credentialing.Regery(hby=self.hby, name=self.hby.name, base=self.hby.base, temp=self.hby.temp)
        )
        self.oobiery: Oobiery = oobiery
        self.load_url = load_url  # Function to load URLs, can be mocked for testing

        super(UniversalResolverResource, self).__init__()

    def on_get(self, req: falcon.Request, rep: falcon.Response, did: str):
        """
        Handle GET requests to resolve a DID by its identifier (KERI AID).

        Parameters:
            req (falcon.Request): The HTTP request object.
            rep (falcon.Response): The HTTP response object.
            did (str): The DID to resolve.
        """
        if did is None or did == '':
            logger.error(f'Missing DID: {did}')
            rep.status = falcon.HTTP_400
            rep.content_type = 'application/json'
            rep.media = {'error': "invalid resolution request body, 'did' is required"}
            return

        try:
            did = didding.requote(did)  # Re-quote the DID to ensure it is valid for resolution
        except ValueError as e:
            logger.error(f'Invalid DID: {did}')
            rep.status = falcon.HTTP_400
            rep.content_type = 'application/json'
            rep.media = {'message': f'invalid DID: {did}', 'error': str(e)}
            return
        logger.info(f'Request to resolve did: {did}')

        if 'oobi' in req.params:
            oobi = req.params['oobi']
            logger.info(f'From parameters {req.params} got oobi: {oobi}')
        else:
            oobi = None

        accepts = req.get_header('Accept')
        accepts = accepts.lower() if accepts else ''

        if 'meta' in req.params:
            meta = req.params['meta'].lower() in ('true', '1', 'yes')
            logger.info(f'From parameters {req.params} got meta: {meta}')
        elif accepts == 'application/did-resolution':
            meta = True
            logger.info(f'From accept header Accept got {accepts}')
        else:
            meta = False

        if did.startswith('did:webs'):
            domain, port, path, aid, query = didding.parse_did_webs(did=did)
            query_vars = didding.parse_query_string(query)
            if 'meta' in query_vars:
                meta = query_vars['meta']
            result, data = resolve(
                hby=self.hby, rgy=self.rgy, did=did, meta=meta, load_url=self.load_url, timeout=self.TimeoutArtifactResolution
            )
        elif did.startswith('did:keri'):
            result, data = resolve_did_keri(self.hby, self.rgy, did, oobi, meta)
        else:
            logger.error(f'Failed to resolve invalid DID: {did}')
            rep.status = falcon.HTTP_400
            rep.media = {'error': f'invalid DID: {did}'}
            return

        if not result:
            logger.error(f'Failed to resolve DID {did}')
            logger.debug(f'Resolution data: {data}')
            rep.status = falcon.HTTP_417
            rep.media = {'error': f'failed to resolve DID: {did}'}
            return

        logger.info(f'Successfully resolved {did}')
        rep.status = falcon.HTTP_200

        if didding.DD_META_FIELD in data:  # meaning, resolution is a DID resolution result and has metadata
            # application/did-resolution is expected by the HTTP Binding in the DID Resolution spec and the Universal Resolver
            rep.set_header('Content-Type', 'application/did-resolution')
        else:
            rep.set_header('Content-Type', 'application/did+ld+json')
        rep.media = data
        return


def resolve_did_keri(hby: habbing.Habery, rgy: credentialing.Regery, did: str, oobi: str = None, meta: bool = False):
    """
    Performs a did:keri DID document resolution based on the KEL retrieved from an OOBI resolution.

    Parameters:
        hby (Habery): identifier database environment
        rgy (Regery): Credential and registry data manager
        did (str): The did:keri DID to resolve.
        oobi (str): OOBI to use for resolution, if None then the AID must be known in the KERI database.
        meta (bool): Whether to include metadata in the DID document.
    Returns:
        tuple(bool, dict): (verified, resolution) where verified is a boolean indicating verification status
    """
    aid, query = didding.parse_did_keri(did)
    if oobi is None and hby.kevers.get(aid) is None:
        return False, {'error': f'Unknown AID, cannot resolve DID {did}'}
    if hby.kevers.get(aid) is not None:  # return early if AID is known
        return True, didding.generate_did_doc(hby, rgy, did=did, aid=aid, meta=meta)
    oobiery = oobiing.Oobiery(hby=hby)
    doist = doing.Doist(limit=10.0, tock=0.03125, real=True)
    deeds = doist.enter(oobiery.doers)

    # Add OOBI record to Baser.oobis so it will be resolved
    obr = basing.OobiRecord(date=helping.nowIso8601())
    obr.cid = aid
    hby.db.oobis.pin(keys=(oobi,), val=obr)
    while (
        hby.db.roobi.get(keys=(oobi,)) is None or aid not in hby.kevers
    ):  # wait for aid in hby.kevers waits for delegates to have their AES found
        doist.recur(deeds)
        hby.kvy.processEscrows()  # for delegated AIDs so authorizing event seals (AES) from delegator ixn evts are found

    return True, didding.generate_did_doc(hby, rgy, did=did, aid=aid, meta=meta)



import json

import pytest
from hio.base import doing

from dws import ArtifactResolveError
from dws.core import requesting


def test_load_url_with_hio():
    url = 'https://weboftrust.github.io/oobi/EN6Oh5XSD5_q2Hgu-aqpdfbVepdpYpFlgz6zvJL5b_r5/'
    resp = requesting.load_url_with_hio_clienter(url)
    assert resp is not None, 'Response should not be None'

    resp = requesting.load_url_with_hio(url)
    assert resp is not None, 'Response should not be None'


def test_load_url_with_hio_invalid_url_throws():
    invalid_url = 'http://invalid-url:123456789:://??##@@'
    with pytest.raises(ValueError) as exc_info:
        requesting.load_url_with_hio(invalid_url)
    assert 'Port could not be cast' in str(exc_info.value)

    with pytest.raises(ValueError) as exc_info:
        requesting.load_url_with_hio_clienter(invalid_url)
    assert 'Port could not be cast' in str(exc_info.value)


def test_create_http_client_with_body_encodes():
    url = 'http://192.0.2.1/api'
    body = json.dumps({'key': 'value'})
    headers = {'Content-Type': 'application/json'}
    client, client_doer = requesting.create_http_client('POST', url, body=body, headers=headers)
    assert client is not None
    assert client.requests[0]['body'] == body.encode('utf-8'), 'Body should be encoded to bytes'

    doist = doing.Doist(limit=0.1, tock=0.03125, real=True)
    clienter = requesting.HTTPClienter()
    clienter.always = False
    clienter.tymth = doist.tymen()  # share the Doist tymth with the Clienter
    client = clienter.request('POST', url, body=body, headers=headers)
    assert client is not None
    assert client.requests[0]['body'] == body.encode('utf-8'), 'Body should be encoded to bytes'


def test_load_url_with_hio_timeout_causes_hio_error():
    # Use RFC 5737 TEST-NET address (192.0.2.0/24) - not routed, connection hangs.
    # example.com responds quickly (404), causing intermittent test failure.
    bad_url = 'http://192.0.2.1/oobi/EN6Oh5XSD5_q2Hgu-aqpdfbVepdpYpFlgz6zvJL5b_r5/'
    with pytest.raises(ArtifactResolveError) as exc_info:
        requesting.load_url_with_hio(bad_url, timeout=0.1)
    assert 'Failed to load URL' in str(exc_info.value), 'Should raise HioError on timeout'

    # requesting.load_url_with_hio_clienter(bad_url, timeout=1.0)
    with pytest.raises(ArtifactResolveError) as exc_info:
        requesting.load_url_with_hio_clienter(bad_url, timeout=0.1)
    assert 'Failed to load URL' in str(exc_info.value), 'Should raise HioError on timeout'



__version__ = '0.3.5'  # also change in pyproject.toml and Makefile

# Logging config
import logging

from hio.help import ogling

from dws.app.logs import TruncatedFormatter

log_name = 'dws'  # name of this project that shows up in log messages
log_format_str = f'%(asctime)s [{log_name}] %(levelname)-8s %(module)s.%(funcName)s-%(lineno)s %(message)s'

ogler = ogling.initOgler(prefix=log_name, syslogged=False)
ogler.level = logging.INFO
formatter = TruncatedFormatter(log_format_str)
formatter.default_msec_format = None
ogler.baseConsoleHandler.setFormatter(formatter)
ogler.reopen(name=log_name, temp=True, clear=True)


def set_log_level(loglevel, logger):
    """Set the log level for the logger."""
    ogler.level = logging.getLevelName(loglevel.upper())
    logger.setLevel(ogler.level)


class DidWebsError(Exception):
    """Base class for all exceptions raised by the dws.app.cli.commands.did.webs module."""

    pass


class ArtifactResolveError(DidWebsError):
    pass


class UnknownAID(DidWebsError):
    """Exception raised when an unknown AID is encountered."""

    def __init__(self, aid: str, did: str):
        super().__init__(f'Unknown AID {aid} found in {did}')
        self.aid = aid



import base64
import json
from typing import List, Optional

import pysodium
from didcomm.common.resolvers import SecretsResolver
from didcomm.common.types import DID, VerificationMaterial, VerificationMaterialFormat, VerificationMethodType
from didcomm.did_doc.did_doc import DIDCommService, DIDDoc, VerificationMethod
from didcomm.did_doc.did_resolver import DIDResolver
from didcomm.secrets.secrets_resolver_demo import Secret
from keri.core import coring, eventing, signing

"""
Utilities for DIDComm packing and unpacking using did:keri as alternative to Peer DID
- Use SICPA didcomm-python library
- Authcryypt message only
- AID is Ed25519 and derive X25519 key pair from same private key
- Transferable AID but with no next key that makes it non transferable (no key rotations)
"""


def createKeriDid():
    salt = signing.Salter()
    signerEd25519 = salt.signer(transferable=True, temp=True)

    X25519_pubkey = pysodium.crypto_sign_pk_to_box_pk(signerEd25519.verfer.raw)
    # Manual CESR coding
    X25519_pubkey_qb64 = 'C' + (base64.urlsafe_b64encode(bytes([0]) + X25519_pubkey).decode('utf-8'))[1:]
    # Or using karipy
    X25519_pubkey_qb64 = coring.Matter(raw=X25519_pubkey, code=coring.MtrDex.X25519).qb64

    serder = eventing.incept(
        keys=[signerEd25519.verfer.qb64],
        data=[{'e': X25519_pubkey_qb64}, {'se': 'https://example.coom/'}],
        code=coring.MtrDex.Blake3_256,  # code is for self-addressing
    )

    did = 'did:keri:' + serder.ked['i']
    kelb64 = base64.urlsafe_b64encode(bytes(json.dumps(serder.ked), 'utf-8')).decode('utf-8')
    long_did = did + '?icp=' + kelb64

    # pp(serder.ked)

    return {'did': did, 'long_did': long_did, 'signer': signerEd25519}


def validateLongDid(long_did):
    # TODO validate URL and make parsing safer
    did = long_did.split('?')[0]
    kelb64 = long_did.split('=')[1] + '=='
    kel_decoded = json.loads(base64.urlsafe_b64decode(kelb64))
    prefixer = coring.Prefixer(ked=kel_decoded)
    return prefixer.qb64b.decode('utf-8') == did.split(':')[2]


class SecretsResolverInMemory(SecretsResolver):
    def __init__(self, store: dict):
        self._store = store

    async def get_key(self, kid: str) -> Optional[Secret]:
        did, kident = kid.split('#')

        signer = self._store[did]['signer']
        X25519_pubkey = pysodium.crypto_sign_pk_to_box_pk(signer.verfer.raw)
        X25519_pubkey_b64 = base64.urlsafe_b64encode(X25519_pubkey).decode('utf-8')
        X25519_prikey = pysodium.crypto_sign_sk_to_box_sk(signer.raw + signer.verfer.raw)
        X25519_prikey_b64 = base64.urlsafe_b64encode(X25519_prikey).decode('utf-8')

        Ed25519_pubkey_raw = signer.verfer.raw
        Ed25519_pubkey_b64 = base64.urlsafe_b64encode(Ed25519_pubkey_raw).decode('utf-8')

        Ed25519_prikey_raw = signer.raw
        Ed25519_prikey_b64 = base64.urlsafe_b64encode(Ed25519_prikey_raw).decode('utf-8')

        secret = Secret(
            kid=kid,
            type=VerificationMethodType.JSON_WEB_KEY_2020,
            verification_material=VerificationMaterial(
                format=VerificationMaterialFormat.JWK,
                value=json.dumps(
                    {'kty': 'OKP', 'crv': 'X25519', 'd': X25519_prikey_b64, 'x': X25519_pubkey_b64, 'kid': kid}
                    if kident == 'key-1'
                    else {'kty': 'OKP', 'crv': 'Ed25519', 'd': Ed25519_prikey_b64, 'x': Ed25519_pubkey_b64, 'kid': kid}
                ),
            ),
        )
        return secret

    async def get_keys(self, kids: List[str]) -> List[str]:
        return kids


class DidKeriResolver(DIDResolver):
    def __init__(self, store: dict):
        self._store = store

    async def resolve(self, did: DID) -> DIDDoc:
        # TODO validate URL and make parsing safer
        short_did = did.split('?')[0]
        if len(did.split('=')) > 1:
            kelb64 = did.split('=')[1] + '=='
            ked = json.loads(base64.urlsafe_b64decode(kelb64))
            self._store[short_did]['ked'] = ked
        else:
            ked = self._store[short_did]['ked']

        # Manual CESR Decoding
        ed25519_pubkey_qb64 = ked['k'][0]
        ed25519_pubkey_b64 = 'A' + ed25519_pubkey_qb64[1:]
        ed25519_pubkey_raw = base64.urlsafe_b64decode(ed25519_pubkey_b64)[1:]
        x1 = base64.urlsafe_b64encode(ed25519_pubkey_raw).decode('utf-8')
        # Or using keripy:
        x1 = base64.urlsafe_b64encode(coring.Matter(qb64=ked['k'][0]).raw).decode('utf-8')

        x2 = base64.urlsafe_b64encode(coring.Matter(qb64=ked['a'][0]['e']).raw).decode('utf-8')

        return DIDDoc(
            did=did,
            key_agreement_kids=[short_did + '#key-1'],
            authentication_kids=[short_did + '#key-2'],
            verification_methods=[
                VerificationMethod(
                    id=short_did + '#key-1',
                    type=VerificationMethodType.JSON_WEB_KEY_2020,
                    controller=did,
                    verification_material=VerificationMaterial(
                        format=VerificationMaterialFormat.JWK, value=json.dumps({'kty': 'OKP', 'crv': 'X25519', 'x': x2})
                    ),
                ),
                VerificationMethod(
                    id=short_did + '#key-2',
                    type=VerificationMethodType.JSON_WEB_KEY_2020,
                    controller=did,
                    verification_material=VerificationMaterial(
                        format=VerificationMaterialFormat.JWK, value=json.dumps({'kty': 'OKP', 'crv': 'Ed25519', 'x': x1})
                    ),
                ),
            ],
            didcomm_services=[
                DIDCommService(id='endpoint-1', service_endpoint=ked['a'][1]['se'], routing_keys=[], accept=['didcomm/v2'])
            ],
        )



import falcon
import pytest
from keri import kering
from keri.app import habbing
from keri.core import eventing
from keri.db import basing, koming, subing
from keri.vdr import credentialing
from mockito import mock, when

from dws.core import didding
from dws.core.artifacting import gen_loc_schemes_cesr
from dws.core.ends.keri_cesr_resource_end import KeriCesrResourceEnd


def test_keri_cesr_resource_end_on_get_single_sig():
    req = mock(falcon.Request)
    rep = mock(falcon.Response)
    hby = mock(habbing.Habery)

    hab = mock()
    hab.kever = mock()
    kever = mock(eventing.Kever)
    kever.wits = ['BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha']
    hab.kevers = {'test_aid': kever}
    # hab.kever.wits = ['BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha']

    hby.db = mock()
    hby.db.ends = mock()
    hby.name = 'test_hab'
    hby.base = 'test_base'
    hby.habs = {'test_aid': hab}
    hby.kvy = mock()
    hby.kevers = {'test_aid': mock()}

    hab.db = mock(basing.Baser)
    hab.db.ends = mock(koming.Komer)

    rgy = mock(credentialing.Regery)
    rgy.reger = mock(credentialing.Reger)
    rgy.reger.issus = mock(subing.CesrDupSuber)
    rgy.reger.schms = mock(subing.CesrDupSuber)
    when(rgy.reger.issus).get(keys='test_aid').thenReturn([])
    when(rgy.reger.schms).get(keys=didding.DES_ALIASES_SCHEMA.encode()).thenReturn([])

    req.path = '/test_aid/keri.cesr'

    mock_serder_raw = bytearray(
        b'{'
        b'"v":"KERI10JSON0001b7_","t":"icp",'
        b'"d":"EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4",'
        b'"i":"EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4",'
        b'"s":"0",'
        b'"kt":"1","k":["DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K"],'
        b'"nt":"1","n":["EHlpcaxffvtcpoUUMTc6tpqAVtb2qnOYVk_3HRsZ34PH"],'
        b'"bt":"2","b":["BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","BLskRTInXnMxWaGqcpSyMgo0nYbalW99cGZESrz3zapM","BIKKuvBwpmDVA4Ds-EpL5bt9OqPzWPja2LigFYZN2YfX"],'
        b'"c":[],"a":[]}'
        b'-VBq-AABAABpDqK4TQytiWO_S-2VvfigdPs6T2pEWPfbgqy7DzYhakD9EmW-wGGa7i5VoF7Re8pkCCLIAO35_BtZOfNV4WIA-BADAADBrKDUOPHm9IFvg_EeEmMMzAvXB4xu6MdnzTohJkeK3Ome__5IWtnWZmXRYyIYau5BPqVXM9RptPc2DCmDg2wKABDrSZ3pVsK7DNlSS_fcT3QO3adZyhcIxcWiJUc5dYsHlEu-A3AVu8nkqXLeYXqE9Z_JKTJen-GfHU3tVp16GPIEACDUzCmXCwY-E6bCbz7umsvnvBS2MS83-03CbCuZ3DZN1GQLlH-A3bUKlhabdqjYW56JtifgcljgGvN7mJk8oa8P-EAB0AAAAAAAAAAAAAAAAAAAAAAA1AAG2025-03-18T17c57c24d927822p00c00'
    )

    # when(hby.db).clonePreIter(pre='test_aid').thenReturn(mock_serder_raw)
    when(hab).replay(pre='test_aid').thenReturn(mock_serder_raw)

    when(hab).loadLocScheme(eid='BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha', scheme='').thenReturn(
        bytearray(
            b'{'
            b'"v":"KERI10JSON0000fa_","t":"rpy",'
            b'"d":"ELSHJwBjsy41VvikaWd5cSC5hoooONVeVsImCzjzBQWP",'
            b'"dt":"2022-01-20T12:57:59.823350+00:00",'
            b'"r":"/loc/scheme",'
            b'"a":{"eid":"BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","scheme":"http","url":"http://127.0.0.1:5642/"}}'
            b'-VAi-CABBBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha0BDRY23xEjEXaAJs7jPJt6uyxK8N2apNJitvn9mo0q4Gh8p7Pf2bAEp1Ufed5l0FdlLxV-Z2sMO8D7wVtA-m_QEM'
            b'{'
            b'"v":"KERI10JSON0000f8_","t":"rpy",'
            b'"d":"EBwDJvb5oW2SgwNfDK8Ib-NiljgBt4uK1bDjW3QztBPr",'
            b'"dt":"2022-01-20T12:57:59.823350+00:00",'
            b'"r":"/loc/scheme",'
            b'"a":{"eid":"BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","scheme":"tcp","url":"tcp://127.0.0.1:5632/"}}'
            b'-VAi-CABBBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha0BCsK_YJIDH8djf5ncLs0VPJ1In104Hiu1392AlIMVFhmIxDP6gxgzMtklcOIyhQwRe7Mvgjniynjdv95iTCPWEL'
        )
    )
    when(hab).loadEndRole(
        cid='BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha',
        eid='BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha',
        role=kering.Roles.controller,
    ).thenReturn(
        bytearray(
            b'{'
            b'"v":"KERI10JSON000113_","t":"rpy",'
            b'"d":"EC-u2taS5Z0YZT18XvV8cPnFDJVNjm6B7j_vZbeMBKHF",'
            b'"dt":"2025-03-21T14:33:26.196052+00:00",'
            b'"r":"/end/role/add",'
            b'"a":{"cid":"EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4","role":"witness","eid":"BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha"}}'
            b'-VA0-FABEKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk40AAAAAAAAAAAAAAAAAAAAAAAEKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4-AABAAAsb68BJ6XGB77xP37tPiDjZOd6oB4nxshznxz6GMy1dTmvpi5yltfvTpBNLZQYlhpRzUI3K0GD_4DNTiUldHAL'
        )
    )
    when(hab.db.ends).getItemIter(keys=('test_aid', kering.Roles.agent)).thenReturn([])

    when(hab.db.ends).getItemIter(keys=('test_aid', kering.Roles.mailbox)).thenReturn(
        [((None, 'mailbox', 'BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'), None)]
    )
    when(hab).loadLocScheme(eid='BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha').thenReturn(
        bytearray(
            b'{"r":"/loc/scheme","a":{"eid":"BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","scheme":"tcp","url":"tcp://127.0.0.1:5632/"}}'
        )
    )
    when(hab).loadEndRole(
        cid='test_aid', eid='BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha', role=kering.Roles.mailbox
    ).thenReturn(
        bytearray(
            b'{"r":"/end/role/add","a":{"eid":"BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","role":"mailbox","cid":"test_aid"}}'
        )
    )

    resource = KeriCesrResourceEnd(hby, rgy)
    resource.on_get(req, rep, 'test_aid')

    assert rep.status == falcon.HTTP_200
    assert rep.content_type == 'application/cesr'
    assert (
        rep.data == b'{'
        b'"v":"KERI10JSON0001b7_","t":"icp",'
        b'"d":"EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4",'
        b'"i":"EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4",'
        b'"s":"0",'
        b'"kt":"1","k":["DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K"],'
        b'"nt":"1","n":["EHlpcaxffvtcpoUUMTc6tpqAVtb2qnOYVk_3HRsZ34PH"],'
        b'"bt":"2","b":["BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","BLskRTInXnMxWaGqcpSyMgo0nYbalW99cGZESrz3zapM","BIKKuvBwpmDVA4Ds-EpL5bt9OqPzWPja2LigFYZN2YfX"],'
        b'"c":[],"a":[]}'
        b'-VBq-AABAABpDqK4TQytiWO_S-2VvfigdPs6T2pEWPfbgqy7DzYhakD9EmW-wGGa7i5VoF7Re8pkCCLIAO35_BtZOfNV4WIA-BADAADBrKDUOPHm9IFvg_EeEmMMzAvXB4xu6MdnzTohJkeK3Ome__5IWtnWZmXRYyIYau5BPqVXM9RptPc2DCmDg2wKABDrSZ3pVsK7DNlSS_fcT3QO3adZyhcIxcWiJUc5dYsHlEu-A3AVu8nkqXLeYXqE9Z_JKTJen-GfHU3tVp16GPIEACDUzCmXCwY-E6bCbz7umsvnvBS2MS83-03CbCuZ3DZN1GQLlH-A3bUKlhabdqjYW56JtifgcljgGvN7mJk8oa8P-EAB0AAAAAAAAAAAAAAAAAAAAAAA1AAG2025-03-18T17c57c24d927822p00c00'
        b'{"v":"KERI10JSON0000fa_","t":"rpy",'
        b'"d":"ELSHJwBjsy41VvikaWd5cSC5hoooONVeVsImCzjzBQWP",'
        b'"dt":"2022-01-20T12:57:59.823350+00:00",'
        b'"r":"/loc/scheme",'
        b'"a":{"eid":"BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","scheme":"http","url":"http://127.0.0.1:5642/"}}'
        b'-VAi-CABBBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha0BDRY23xEjEXaAJs7jPJt6uyxK8N2apNJitvn9mo0q4Gh8p7Pf2bAEp1Ufed5l0FdlLxV-Z2sMO8D7wVtA-m_QEM'
        b'{"v":"KERI10JSON0000f8_","t":"rpy",'
        b'"d":"EBwDJvb5oW2SgwNfDK8Ib-NiljgBt4uK1bDjW3QztBPr",'
        b'"dt":"2022-01-20T12:57:59.823350+00:00",'
        b'"r":"/loc/scheme",'
        b'"a":{"eid":"BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","scheme":"tcp","url":"tcp://127.0.0.1:5632/"}}'
        b'-VAi-CABBBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha0BCsK_YJIDH8djf5ncLs0VPJ1In104Hiu1392AlIMVFhmIxDP6gxgzMtklcOIyhQwRe7Mvgjniynjdv95iTCPWEL'
        b'{"v":"KERI10JSON000113_","t":"rpy",'
        b'"d":"EC-u2taS5Z0YZT18XvV8cPnFDJVNjm6B7j_vZbeMBKHF",'
        b'"dt":"2025-03-21T14:33:26.196052+00:00",'
        b'"r":"/end/role/add",'
        b'"a":{"cid":"EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4","role":"witness","eid":"BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha"}}'
        b'-VA0-FABEKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk40AAAAAAAAAAAAAAAAAAAAAAAEKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4-AABAAAsb68BJ6XGB77xP37tPiDjZOd6oB4nxshznxz6GMy1dTmvpi5yltfvTpBNLZQYlhpRzUI3K0GD_4DNTiUldHAL'
        b'{"r":"/loc/scheme","a":{"eid":"BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","scheme":"tcp","url":"tcp://127.0.0.1:5632/"}}'
        b'{"r":"/end/role/add","a":{"eid":"BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","role":"mailbox","cid":"test_aid"}}'
    )


def test_gen_loc_schemes_cesr_with_mailbox_role_skips_agent():
    hab = mock()
    hab.kever = mock()
    kever = mock(eventing.Kever)
    hab.kevers = {'test_aid': kever}
    # hab.kever.wits = ['BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha']

    hab.db = mock(basing.Baser)
    hab.db.ends = mock(koming.Komer)

    when(hab.db.ends).getItemIter(keys=('test_aid', kering.Roles.mailbox)).thenReturn(
        [((None, 'mailbox', 'BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'), None)]
    )

    when(hab).loadLocScheme(eid='BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha').thenReturn(
        bytearray(
            b'{"r":"/loc/scheme","a":{"eid":"BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","scheme":"tcp","url":"tcp://127.0.0.1:5632/"}}'
        )
    )

    when(hab).loadEndRole(
        cid='test_aid', eid='BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha', role=kering.Roles.mailbox
    ).thenReturn(
        bytearray(
            b'{"r":"/end/role/add","a":{"eid":"BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","role":"mailbox","cid":"test_aid"}}'
        )
    )
    loc_scheme_cesr = gen_loc_schemes_cesr(hab, 'test_aid', role=kering.Roles.mailbox)
    assert (
        loc_scheme_cesr
        == b'{"r":"/loc/scheme","a":{"eid":"BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","scheme":"tcp","url":"tcp://127.0.0.1:5632/"}}'
        b'{"r":"/end/role/add","a":{"eid":"BDilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha","role":"mailbox","cid":"test_aid"}}'
    )


def test_keri_cesr_resource_end_on_get_bad_path():
    req = mock(falcon.Request)
    rep = mock(falcon.Response)
    hby = mock(habbing.Habery)
    hby.name = 'test_hab'
    hby.base = 'test_base'
    hby.kvy = mock()
    hby.db = mock()
    hby.kevers = {'test_aid': mock()}

    rgy = mock(credentialing.Regery)

    req.path = '/test_aid/bad.path'

    resource = KeriCesrResourceEnd(hby, rgy)

    with pytest.raises(falcon.HTTPBadRequest) as e:
        resource.on_get(req, rep, 'test_aid')

    assert isinstance(e.value, falcon.HTTPBadRequest)


def test_keri_cesr_resource_end_on_get_bad_aid():
    req = mock(falcon.Request)
    rep = mock(falcon.Response)
    hby = mock(habbing.Habery)
    hby.name = 'test_hab'
    hby.base = 'test_base'
    hby.kvy = mock()
    hby.db = mock()
    hby.kevers = {'test_aid': mock()}

    rgy = mock(credentialing.Regery)

    req.path = '/bad_aid/keri.cesr'

    resource = KeriCesrResourceEnd(hby, rgy)

    with pytest.raises(falcon.HTTPNotFound) as e:
        resource.on_get(req, rep, 'bad_aid')

    assert isinstance(e.value, falcon.HTTPNotFound)



# -*- encoding: utf-8 -*-
"""
dws.app.cli.commands module

"""

import multicommand
from keri.app import directing

from dws.app.cli import commands


def main():
    """Run the dws CLI app."""
    parser = multicommand.create_parser(commands)
    args = parser.parse_args()

    if not hasattr(args, 'handler'):
        parser.print_help()
        return 0

    try:
        doers = args.handler(args)
        directing.runController(doers=doers, expire=0.0)
        return 0
    except Exception as ex:
        import os

        if os.getenv('DEBUG_DWS'):
            import traceback

            traceback.print_exc()
        else:
            print(f'ERR: {ex}')
        return -1


if __name__ == '__main__':
    main()



import pytest
from keri import kering
from keri.app import habbing
from keri.app.habbing import Habery
from keri.core import eventing
from keri.db import basing, koming
from keri.vdr.credentialing import Regery
from mockito import mock, when

from dws import DidWebsError
from dws.core import artifacting


def test_make_keri_cesr_path_with_nonexistent_dir_creates_path():
    """
    Test that make_keri_cesr_path creates the necessary directories when the output directory does not exist.
    """
    import os
    import tempfile

    with tempfile.TemporaryDirectory() as temp_dir:
        aid = 'test_aid'
        path = artifacting.make_keri_cesr_path(temp_dir, aid)
        assert os.path.exists(os.path.dirname(path))
        assert path == os.path.join(temp_dir, aid)  # Check the full path


def test_make_did_json_path_with_nonexistent_dir_creates_path():
    """
    Test that make_did_json_path creates the necessary directories when the output directory does not exist.
    """
    import os
    import tempfile

    with tempfile.TemporaryDirectory() as temp_dir:
        aid = 'test_aid'
        path = artifacting.make_did_json_path(temp_dir, aid)
        assert os.path.exists(os.path.dirname(path))
        assert path == os.path.join(temp_dir, aid)  # Check the full path


def test_generate_artifacts_no_diddoc_raises_didwebs_error():
    """
    Test that generate_artifacts raises DidWebsError when the DID document generation fails.
    """

    hby = Habery(name='test_hab', base='test_base')
    rgy = Regery(hby=hby, name='test_regery')

    did = 'did:webs:example.com%3A1234:test_path:EJg2UL2kSzFV_Akd9ISAvgOUFDKcBxpDO3OZDIbSIjGe'

    with pytest.raises(DidWebsError):
        artifacting.generate_artifacts(hby, rgy, did, meta=True, output_dir='./tests/artifact_output_dir')


def test_gen_loc_schemes_cesr_for_agent_returns_loc_scheme_and_endrole():
    aid = 'test_aid'
    eid = 'test_eid'
    role = kering.Roles.agent
    hab = mock(habbing.Hab)
    kever = mock(eventing.Kever)
    hab.kevers = {aid: kever}
    hab.db = mock(basing.Baser)
    hab.db.ends = mock(koming.Komer)
    when(hab.db.ends).getItemIter(keys=(aid, kering.Roles.agent)).thenReturn([((None, kering.Roles.agent, eid), None)])
    when(hab).loadLocScheme(eid=eid, scheme='').thenReturn(bytearray(b'loc_scheme_data'))
    when(hab).loadEndRole(cid=aid, eid=eid, role=kering.Roles.agent).thenReturn(bytearray(b'end_role_data'))

    msgs = artifacting.gen_loc_schemes_cesr(hab, aid, role=role)
    assert msgs == bytearray(b'loc_scheme_dataend_role_data')



import json
import os

import falcon
from keri.app import habbing
from keri.vdr import credentialing

from dws.core import didding

DID_JSON = 'did.json'


class DIDWebsResourceEnd:
    """
    did.json HTTP resource for accessing did:webs DID documents for KERI AIDs.
    """

    def __init__(self, hby: habbing.Habery, rgy: credentialing.Regery, meta: bool = False):
        """
        Initialize did:webs did.json artifact endpoint that will pull designated aliases from the specified registry
        and will optionally include metadata in the DID document.

        Parameters:
            hby (Habery): Database environment for AIDs to expose
            rgy (Regery): Registry for credential and registry data
            meta (bool): Whether to include metadata in the DID document. Default is False.
        """
        self.hby = hby
        self.rgy = rgy
        self.meta = meta
        super().__init__()

    def on_get(self, req, rep, aid):
        """GET endpoint for resolving KERI AIDs as did:web DIDs

        Parameters:
            req (Request) Falcon HTTP Request object:
            rep (Response) Falcon HTTP Response object:
            aid (str): AID to resolve, or path used if None
        """
        # Read the DID from the parameter extracted from path or manually extract
        if not req.path.endswith(f'/{DID_JSON}'):
            raise falcon.HTTPBadRequest(description=f'invalid did:web DID URL {req.path}')

        # 404 if AID not recognized
        if aid not in self.hby.kevers:
            raise falcon.HTTPNotFound(description=f'KERI AID {aid} not found')

        path = os.path.normpath(req.path).replace(f'/{DID_JSON}', '').replace('/', ':')
        port = req.get_header('X-Forwarded-Port') or req.port  # support for reverse proxy forwarding
        port_str = '' if port in ('80', '443') else f'%3A{port}'

        did = f'did:web:{req.host}{port_str}{path}'

        # Generate the DID Doc and return
        diddoc = didding.generate_did_doc(self.hby, self.rgy, did, aid, meta=self.meta)

        rep.status = falcon.HTTP_200
        rep.content_type = 'application/json'
        rep.data = json.dumps(diddoc, indent=2).encode('utf-8')



# -*- encoding: utf-8 -*-
"""
dws.core.webbing module

"""

import falcon
from keri.app import habbing
from keri.vdr import credentialing

from dws.core import ends


def load_endpoints(
    app: falcon.App, hby: habbing.Habery, rgy: credentialing.Regery, did_path: str = '', meta: bool = False
) -> None:
    """
    Set up web app endpoints to serve configured KERI AIDs as `did:web` DIDs

    Parameters:
        app (App): Falcon app to register endpoints against
        hby (Habery): Database environment for exposed KERI AIDs
        did_path (str): Optional prefixed path segment to include in the URL for did:webs. Defaults to empty string.
        meta (bool): Whether to include metadata in the DID document. Default is False.
    """
    app.add_route('/health', ends.HealthEnd())
    did_webs_path = '' if not did_path else f'/{did_path}'
    app.add_route(f'{did_webs_path}/{{aid}}/did.json', ends.DIDWebsResourceEnd(hby, rgy, meta=meta))
    app.add_route(f'{did_webs_path}/{{aid}}/keri.cesr', ends.KeriCesrResourceEnd(hby, rgy))



import io
import json
import logging
import os
from unittest.mock import Mock, patch

import falcon
import pytest
from hio.base import doing
from keri import core
from keri.app import configing, habbing, oobiing
from keri.vdr import credentialing
from mockito import mock

from dws.core import artifacting, didding, generating, resolving
from dws.core.ends import DID_JSON, KERI_CESR
from dws.core.resolving import RequestLoggerMiddleware
from tests import conftest
from tests.conftest import CredentialHelpers, HabbingHelpers, WitnessContext, self_attested_aliases_cred_subj


def test_artifact_generation_creates_expected_artifacts():
    salt = b'0AAB_Fidf5WeZf6VFc53IxVw'
    registry_nonce = '0ADV24br-aaezyRTB-oUsZJE'

    # set up witness config
    wit_salt = core.Salter(raw=b'abcdef0123456789').qb64
    wit_cf = configing.Configer(name='wil', temp=False, reopen=True, clear=False)
    wit_cf_dict = {
        'dt': '2022-01-20T12:57:59.823350+00:00',
        'wil': {'dt': '2022-01-20T12:57:59.823350+00:00', 'curls': ['tcp://127.0.0.1:6633/', 'http://127.0.0.1:6643/']},
    }
    wit_cf.put(wit_cf_dict)
    wil_oobi = 'http://127.0.0.1:6643/oobi/BKH72-JtJiyIgvFA9YAZW4FCrYwICboCb6STGzJwFEv0/controller?name=Wil&tag=witness'

    # Config of the AID controller keystore who is having their did:webs or did:keri artifacts resolved
    rb_cf = configing.Configer(name='red_book', temp=False, reopen=True, clear=False)
    rb_cf.put({'dt': '2022-01-20T12:57:59.823350+00:00', 'iurls': [wil_oobi]})

    # Open the witness Habery and Hab, feed it into the witness setup, and then create the AID controller Habery and Hab
    with (
        HabbingHelpers.openHab(salt=bytes(wit_salt, 'utf-8'), name='wil', transferable=False, temp=True, cf=wit_cf) as (
            wit_hby,
            wit_hab,
        ),
        WitnessContext.with_witness(name='wil', hby=wit_hby, tcp_port=6633, http_port=6643) as wil_wit,
        habbing.openHab(salt=salt, name='red_book', transferable=True, temp=True, cf=rb_cf) as (rb_hby, rb_hab),
    ):
        wil_pre = 'BKH72-JtJiyIgvFA9YAZW4FCrYwICboCb6STGzJwFEv0'
        tock = 0.03125
        doist = doing.Doist(limit=0.0, tock=tock, real=True)
        # Doers and deeds for witness wil
        wit_deeds = doist.enter(doers=wil_wit.doers)

        # Resolve OOBI
        oobiery = oobiing.Oobiery(hby=rb_hby)
        authn = oobiing.Authenticator(hby=rb_hby)
        oobiery_deeds = doist.enter(doers=oobiery.doers + authn.doers)
        while not oobiery.hby.db.roobi.get(keys=(wil_oobi,)):
            doist.recur(deeds=wit_deeds + oobiery_deeds)
            rb_hby.kvy.processEscrows()  # process any escrows from witness receipts
        print(f'Resolved OOBI: {wil_oobi} to {oobiery.hby.db.roobi.get(keys=(wil_oobi,))}')

        # Doers and deeds for the page Hab and Habery
        rb_doers, hby_doer, wit_rcptr_doer = HabbingHelpers.habery_doers(hby=rb_hby)
        rb_deeds = doist.enter(doers=rb_doers)

        page_hab = rb_hby.makeHab(name='page', isith='1', icount=1, toad=1, wits=[wil_pre])

        # Waiting for witness receipts...
        wit_rcptr_doer.msgs.append(dict(pre=page_hab.pre))
        while not wit_rcptr_doer.cues:
            doist.recur(deeds=wit_deeds + rb_deeds)

        # Prepare the DID artifact stream using self-attested credentials
        aid = 'EJg2UL2kSzFV_Akd9ISAvgOUFDKcBxpDO3OZDIbSIjGe'  # page_hab.pre
        host = '127.0.0.1'
        port = f'7677'
        did_path = 'dws'
        meta = True
        did_webs_did = f'did:webs:{host}%3A{port}:{did_path}:{aid}?meta=true'  # did:webs:127.0.0.1%3A7677:dws:EPTCU0Yge3UDk4bsTWtUhiNJjPDZlW88qRQvJpZ53WRV?meta=true

        # Expects the test to be run from the root of the repository
        regery = credentialing.Regery(hby=rb_hby, name=rb_hby.name, temp=rb_hby.temp)
        schema_json = conftest.Schema.designated_aliases_schema()
        rules_json = conftest.Schema.designated_aliases_rules()
        subject_data = self_attested_aliases_cred_subj(host, aid, port, did_path)
        CredentialHelpers.add_cred_to_aid(
            hby=rb_hby,
            hby_doer=hby_doer,
            regery=regery,
            hab=page_hab,
            schema_said='EN6Oh5XSD5_q2Hgu-aqpdfbVepdpYpFlgz6zvJL5b_r5',  # Designated Aliases Public Schema
            schema_json=schema_json,
            subject_data=subject_data,
            rules_json=rules_json,
            recp=None,  # No recipient for self-attested credential
            registry_nonce=registry_nonce,
            additional_deeds=wit_deeds + rb_deeds,
        )

        # get keri.cesr
        reger = regery.reger
        keri_cesr = bytearray()
        # self.retrieve_kel_via_oobi() # not currently used; an alternative to relying on a local KEL keystore
        keri_cesr.extend(artifacting.gen_kel_cesr(rb_hab, aid))  # add KEL CESR stream
        keri_cesr.extend(artifacting.gen_loc_schemes_cesr(rb_hab, aid))
        keri_cesr.extend(artifacting.gen_des_aliases_cesr(page_hab, reger, aid))

        did_webs_diddoc = didding.generate_did_doc(rb_hby, rgy=regery, did=did_webs_did, aid=aid, meta=meta)

        # Run the DIDArtifactGenerator to generate the artifacts
        output_dir = './tests/artifact_output_dir'
        did_art_gen = generating.DIDArtifactGenerator(
            name=rb_hby.name,
            base=rb_hby.base,
            bran=None,
            hby=rb_hby,
            hby_doer=hby_doer,
            regery=regery,
            did=did_webs_did,
            meta=meta,
            output_dir=output_dir,
            verbose=True,
            cf=rb_cf,
        )
        doist.do([did_art_gen])
        assert os.path.exists(f'{output_dir}/{aid}/{DID_JSON}')
        assert os.path.exists(f'{output_dir}/{aid}/{KERI_CESR}')
        assert (
            did_art_gen.did_json[didding.DD_FIELD] == didding.to_did_web(diddoc=did_webs_diddoc, meta=meta)[didding.DD_FIELD]
        ), 'DID document does not match the expected structure'
        assert did_art_gen.keri_cesr == keri_cesr

        with open(f'{output_dir}/{aid}/{DID_JSON}', 'r') as did_file:
            did_json = json.load(did_file)
            assert did_json == didding.to_did_web(diddoc=did_webs_diddoc, meta=meta)

        with open(f'{output_dir}/{aid}/{KERI_CESR}', 'rb') as cesr_file:
            keri_cesr_data = cesr_file.read()
            assert keri_cesr_data == keri_cesr

        assert resolving.diff_dicts(did_webs_diddoc, did_json), 'DID document does not match the expected structure'

        # And test a failed verification by having one be meta=True and one be meta=False

        did_webs_diddoc = didding.generate_did_doc(rb_hby, rgy=regery, did=did_webs_did, aid=aid, meta=True)
        did_webs_diddoc[didding.DD_FIELD] = {f'{didding.VMETH_FIELD}': {}}  # break the diddoc
        result, dd_expected = resolving.verify(did_art_gen.did_json, did_webs_diddoc, meta=True)
        assert not result, 'Verification should fail when meta is True but did document is generated with meta=False'


def test_did_art_genr_with_empty_hby_creates_hby():
    hby_mock = mock(habbing.Habery)
    hby_mock.db = mock()
    hby_doer_mock = mock(habbing.HaberyDoer)
    rgy_mock = mock(credentialing.Regery)
    with patch('dws.core.habs.get_habery_and_doer') as mock_get_habery_and_doer:
        mock_get_habery_and_doer.return_value = hby_mock, hby_doer_mock
        gen = generating.DIDArtifactGenerator(
            name='test_resolver', base='test_base', bran=None, did='fake:did', regery=rgy_mock
        )
        assert gen.hby == hby_mock, 'Expected KeriResolver to create a new Habery instance'
        assert hby_doer_mock in gen.doers, 'Expected KeriResolver to add HaberyDoer to its doers'


@pytest.fixture
def mock_req():
    req = Mock(spec=falcon.Request)
    req.method = 'POST'
    req.url = '/test'
    req.headers = {'Content-Type': 'application/json'}
    req.content_length = 10
    req.stream = io.BytesIO(b'{"key": "value"}')  # Mock stream for body read
    req.env = {'wsgi.input': req.stream, 'CONTENT_LENGTH': '10'}
    return req


@pytest.fixture
def mock_resp():
    return Mock(spec=falcon.Response)


def test_request_logger_middleware_on_debug(mock_req, mock_resp):
    """
    Test that the request logger middleware logs requests when debug is enabled.
    """
    middleware = RequestLoggerMiddleware()

    with (
        patch('dws.core.resolving.logger.isEnabledFor') as mock_enabled,
        patch('dws.core.resolving.logger.debug') as mock_debug,
    ):
        mock_enabled.return_value = True  # Force enabled path

        middleware.process_request(mock_req, mock_resp)

        mock_enabled.assert_called_with(logging.DEBUG)
        mock_debug.assert_called_with('Request body    : %s', '{"key": "value"}')  # Or '<empty>' if no body
        assert mock_req.env['wsgi.input'].read() == b'{"key": "value"}'  # Reset worked
        assert mock_req.env['CONTENT_LENGTH'] == '16'  # Updated to actual len
